from typing import Union
from anndata import AnnData
from mudata import MuData
import numpy as np
from scipy import sparse
from multiprocessing import Pool, cpu_count
import logging

logging.basicConfig(
    format='%(asctime)s %(levelname)-8s %(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S')


def compute_deviations(data: Union[AnnData, MuData], n_jobs=-1):
    """
    Compute raw and bias-corrected deviations.

    Args:
        data (Union[AnnData, MuData]): 
            AnnData object with peak counts or MuData object with 'atac' modality.
        n_jobs:
            Number of cpus used for motif matching. If set to -1, all cpus will be used. Default: -1
    """

    if isinstance(data, AnnData):
        adata = data
    elif isinstance(data, MuData) and "atac" in data.mod:
        adata = data.mod["atac"]
    else:
        raise TypeError(
            "Expected AnnData or MuData object with 'atac' modality")

    # check if the object contains bias in Anndata.varm
    assert "bg_peaks" in adata.varm_keys(
    ), "Cannot find background peaks in the input object, please first run get_bg_peaks!"

    if isinstance(adata.X, sparse.csr_matrix):
        adata.X = np.array(adata.X.todense())

    logging.info('computing expectation reads per cell and peak...')
    expectation = compute_expectation(count=adata.X)

    logging.info('computing observed motif deviations...')
    motif_match = adata.varm['motif_match'].transpose()

    obs_dev = _compute_dev(
        (motif_match, adata.X.transpose(), expectation.transpose())).transpose()

    # compute background deviations for bias-correction
    logging.info('computing background deviations...')

    n_bg_peaks = adata.varm['bg_peaks'].shape[1]
    bg_dev = np.zeros(shape=(n_bg_peaks, adata.n_obs, len(
        adata.uns['motif_name'])), dtype=np.float32)

    if n_jobs == -1:
        n_jobs = cpu_count()

    if n_jobs == 1:
        for i in range(n_bg_peaks):
            bg_peak_idx = adata.varm['bg_peaks'][:, i]
            bg_motif_match = adata.varm['motif_match'][bg_peak_idx, :].transpose()
            bg_dev[i, :, :] = _compute_dev((bg_motif_match, adata.X.transpose(),
                                            expectation.transpose())).transpose()

    elif n_jobs > 1:
        # prepare arguments for multiprocessing
        arguments_list = list()
        for i in range(n_bg_peaks):
            bg_peak_idx = adata.varm['bg_peaks'][:, i]
            bg_motif_match = adata.varm['motif_match'][bg_peak_idx, :].transpose()
            arguments = (bg_motif_match, adata.X.transpose(), expectation.transpose())
            arguments_list.append(arguments)

        # run the function with multiple cpus
        with Pool(processes=n_jobs) as pool:
            all_results = pool.map(_compute_dev, arguments_list)

        # parse the results
        for i in range(n_bg_peaks):
            bg_dev[i, :, :] = all_results[i].transpose()

    mean_bg_dev = np.mean(bg_dev, axis=0)
    std_bg_dev = np.std(bg_dev, axis=0)

    dev = (obs_dev - mean_bg_dev) / std_bg_dev
    dev = np.nan_to_num(dev, 0)

    dev = AnnData(dev, dtype=np.float32)
    dev.obs_names = adata.obs_names
    dev.var_names = adata.uns['motif_name']

    return dev


def _compute_dev(arguments):
    motif_match, count, expectation = arguments

    observed = np.dot(motif_match, count)
    expected = np.dot(motif_match, expectation)

    return ((observed - expected) / expected)


def compute_expectation(count: np.array) -> np.array:
    """
    Compute expetation accessibility per peak and per cell by assuming identical 
    read probability per peak for each cell with a sequencing depth matched to that cell
    observed sequencing depth.
    Args:
        count (_type_): _description_
    """

    a = np.sum(count, axis=0, keepdims=True)
    a /= np.sum(count)

    b = np.sum(count, axis=1, keepdims=True)

    exp = np.dot(b, a)

    return exp



import numpy as np
import scipy as sp
import re

from typing import Union
from anndata import AnnData
from mudata import MuData
from pysam import Fastafile
from tqdm import tqdm
from pynndescent import NNDescent

def get_bg_peaks(data: Union[AnnData, MuData], niterations=50, n_jobs=-1):
    """
    Find background peaks based on GC bias.

    Args:
        data (Union[AnnData, MuData]):
            AnnData object with peak counts or MuData object with 'atac' modality.
        niterations (int, optional): 
            Number of background peaks to sample. Defaults to 50.
        n_jobs:

    Raises:
        TypeError: _description_
    """
    if isinstance(data, AnnData):
        adata = data
    elif isinstance(data, MuData) and "atac" in data.mod:
        adata = data.mod["atac"]
    else:
        raise TypeError(
            "Expected AnnData or MuData object with 'atac' modality")

    # check if the object contains bias in Anndata.varm
    assert "gc_bias" in adata.var.columns, "Cannot find gc bias in the input object, please first run add_gc_bias!"

    reads_per_peak = np.log10(np.sum(adata.X, axis=0))

    # here if reads_per_peak is a numpy matrix, convert it to array
    if isinstance(reads_per_peak, np.matrix):
        reads_per_peak = np.squeeze(np.asarray(reads_per_peak))

    mat = np.array([reads_per_peak, adata.var['gc_bias'].values])
    chol_cov_mat = np.linalg.cholesky(np.cov(mat))
    trans_norm_mat = sp.linalg.solve_triangular(
        a=chol_cov_mat, b=mat, lower=True).transpose()

    index = NNDescent(trans_norm_mat, metric="euclidean", n_neighbors=niterations, n_jobs=n_jobs)
    knn_idx, _ = index.query(trans_norm_mat, niterations)

    adata.varm['bg_peaks'] = knn_idx

    return None


def add_peak_seq(data: Union[AnnData, MuData], genome_file: str, delimiter="-"):
    """
    Add the DNA sequence of each peak to data object. 
    The sequences will be used in GC bias estimation and motif binding sites matching.

    Args:
        data (Union[AnnData, MuData]): 
            AnnData object with peak counts or MuData object with 'atac' modality.
        genome_file (str): 
            Filename of genome reference
        delimiter (str, optional): 
            Delimiter that separates peaks. Defaults to "-".
    """

    if isinstance(data, AnnData):
        adata = data
    elif isinstance(data, MuData) and "atac" in data.mod:
        adata = data.mod["atac"]
    else:
        raise TypeError("Expected AnnData or MuData object with 'atac' modality")

    fasta = Fastafile(genome_file)
    adata.uns['peak_seq'] = [None] * adata.n_vars

    for i in tqdm(range(adata.n_vars)):
        peak = re.split(delimiter, adata.var_names[i])
        chrom, start, end = peak[0], int(peak[1]), int(peak[2])
        adata.uns['peak_seq'][i] = fasta.fetch(chrom, start, end).upper()

    return None


def add_gc_bias(data: Union[AnnData, MuData]):
    """
    Compute GC bias for each peak.

    Args:
        data (Union[AnnData, MuData]): 
            AnnData object with peak counts or MuData object with 'atac' modality.
    Returns:
        _type_: _description_
    """

    if isinstance(data, AnnData):
        adata = data
    elif isinstance(data, MuData) and "atac" in data.mod:
        adata = data.mod["atac"]
    else:
        raise TypeError("Expected AnnData or MuData object with 'atac' modality")

    assert "peak_seq" in adata.uns_keys(), "Cannot find sequences, please first run add_peak_seq!"

    bias = np.zeros(adata.n_vars)

    for i in tqdm(range(adata.n_vars)):
        seq = adata.uns['peak_seq'][i]

        freq_a = seq.count("A")
        freq_c = seq.count("C")
        freq_g = seq.count("G")
        freq_t = seq.count("T")

        if freq_a + freq_c + freq_g + freq_t == 0:
            bias[i] = 0.5
        else:
            bias[i] = (freq_g + freq_c) / (freq_a + freq_c + freq_g + freq_t)

    adata.var['gc_bias'] = bias

    return None


__version__ = "0.0.3"
__version_info__ = tuple([int(num) for num in __version__.split('.')])  # noqa: F401

from .preprocessing import *
from .match_motif import match_motif
from .compute_deviations import compute_deviations
from .get_genome import get_genome




import numpy as np
from pychromvar.compute_deviations import *

def test_compute_expectation():
    count = np.array([[1, 0, 1, ], [0, 1, 1]], dtype=np.float32)
    exp = compute_expectation(count)

    # make sure output has same dimensionas
    assert exp.shape == count.shape

    # check the output
    assert np.array_equal(exp, np.array([[0.5, 0.5, 1], [0.5, 0.5, 1]]))



import os
import wget
import gzip
from tqdm import tqdm

base_url = "https://ftp.ebi.ac.uk/pub/databases/gencode"

genome_url = {
    "hg19": f"{base_url}/Gencode_human/release_42/GRCh37_mapping/GRCh37.primary_assembly.genome.fa.gz",
    "hg38": f"{base_url}/Gencode_human/release_42/GRCh38.primary_assembly.genome.fa.gz",
    "mm10": f"{base_url}/Gencode_mouse/release_M1/NCBIM37.genome.fa.gz",
    "mm39": f"{base_url}/Gencode_mouse/release_M31/GRCm39.primary_assembly.genome.fa.gz"
}

def get_genome(genome:str="hg38", output_dir:str=None):
    """
    Download genome

    Args:
        genome (str, optional): 
            Which genome should be downloaded. Available options are: "hg19", "hg38", "mm9", "mm10".
            Defaults to "hg38".

        output_dir (str):
            Output directory. Default: current directory.
    """
    assert genome in ["hg19", "hg38", "mm10", "mm39"], f"Cannot find {genome}!"

    if not os.path.exists(output_dir):
        output_dir = os.getcwd()

    output_fname = os.path.join(output_dir, f"{genome}.fa")
    if os.path.exists(output_fname):
        os.remove(output_fname)

    gz_fname = os.path.join(output_dir, f"{genome}.fa.gz")
    wget.download(genome_url[genome], gz_fname, bar=bar_thermometer)

    output_fname = os.path.join(output_dir, f"{genome}.fa")
    with open(output_fname, "w") as f:
        gz_file = gzip.open(gz_fname, "rb")
        f.write(gz_file.read().decode("utf-8"))
        gz_file.close()

    os.remove(gz_fname)
